program pi_calculation;
//円周率をgoaldigit桁計算します。32bitの2進数は10進9.63桁に相当 log(2^32)
//並列処理のためWindows専用にコーディングしています
//円周率の計算にマチンの公式の変形を使用

uses
  sysutils, MTProcs;  //このコードはwindows専用

const
  maxdigit = 10000000;  //計算する桁数 1000以上で10の倍数
  maxsize = maxdigit div 8; //計算誤差防止のため配列を20%大きく取る

type
  buftyp = array [0..maxsize] of longword; //32bit 符号なし整数の配列
//配列の添字[0]と[1]の間が小数点位置。固定小数点の2進数として32ビットごとに格納

var
  resbuf : array[1..4] of buftyp; //結果を格納する配列。4つに分割して並列処理するため4つ必要
  goaldigit, arraysize : longint; //計算する桁数と実際に使用する配列の大きさ

procedure clearbuf(var buf: buftyp); //バッファ領域をクリア
var
  i  : integer;
begin
  for i:=0 to arraysize do buf[i]:=0;
end;

procedure divide1(var buf: buftyp; divider: cardinal; var convp: integer);
//bufをdividerで割って結果をbufに格納。convpは収束ポインタ、これより上位はオールゼロ
//割り算は上位から下位に処理
var
  p1,p2  : pointer;
begin
  p1:=@buf[convp-1]; //割られる配列の開始位置
  p2:=@buf[arraysize]; //終了位置
  {$ASMMODE intel}  //高速化のためアセンブラで記述
  asm
    mov ESI,p1 //配列のアドレスをセット
    mov EDX,0 //最初は余りなし
@DIV1LP:
    mov EAX,[ESI]
    div divider
    mov [ESI],EAX //同じ配列に結果を保存
    add ESI,4 //次の配列にポインタを移動(32ビットはアドレスで4進める)
    cmp ESI,p2
    jbe @DIV1LP
  end ['EAX','EDX','ESI']; //アセンブラコードで使用したレジスタをFPC(コンパイラ)に通知
  if buf[convp]=0 then convp:=convp+1; //収束の場合ポインタを進める
end;

procedure divide2(var buf1,buf2:buftyp; divider: cardinal; convp: integer);
//buf1をdividerで割って結果をbuf2に格納。convpは収束ポインタ、これより上位はオールゼロ
//割り算は上位から下位に処理
var
  p1,p2,p3  : pointer;
begin
    p1:=@buf1[convp-1]; //割られる配列開始位置
    p2:=@buf1[arraysize]; //終了位置
    p3:=@buf2[convp-1]; //結果保存配列の開始位置
    {$ASMMODE intel}  //高速化のためアセンブラで記述
    asm
      mov ESI,p1 //元配列のアドレス
      mov EDI,p3 //保存先配列のアドレス
      mov EDX,0 //最初は余りなし
  @DIV1LP:
      mov EAX,[ESI]
      div divider
      mov [EDI],EAX //結果保存配列に保存
      add ESI,4 //次の配列にポインタを移動
      add EDI,4 //保存先配列のポインタを移動
      cmp ESI,p2
      jbe @DIV1LP
    end ['EAX','EDX','ESI','EDI']; //使用したレジスタをFPCに通知
end;

procedure mul100k(var buf: buftyp; arraylimit: integer);
//2進10進変換用に使用する10万倍ルーチン
//10万倍して少数点以上に繰り上がった部分を10進5桁として取る
var
  p1,p2  : pointer;
begin
  p1:=@buf[arraylimit]; //掛けられる配列の開始位置
  p2:=@buf[0]; //終了位置
  {$ASMMODE intel}  //高速化のためアセンブラで記述
  asm
    mov EDI,p1 //ポインタをセット
    mov ECX,0 //最初は桁上げなし
    mov EBX,100000 //10万倍して10進5桁の桁上げ
@MULLP:
    mov EAX,[EDI]
    mul EBX // 掛け算
    add EAX,ECX //前の桁上げ分を足す
    adc EDX,0 //上の足し算の桁上げを次回に繰り越す
    mov [EDI],EAX //結果を保存
    mov ECX,EDX //桁上げ分を保存
    sub EDI,4 //ポインタを移動
    cmp EDI,p2
    jge @MULLP
  end ['EAX','EBX','ECX','EDX','EDI']; //使用したレジスタをFPCに通知
end;


procedure add(var buf1,buf2: buftyp; convp:integer);
//buf1にbuf2を加える。結果はbuf1に格納。convpは収束ポインタ、これより上位はオールゼロ
//足し算は下位から上位に処理
var
  p1,p2,p3  : pointer;
begin
    p1:=@buf1[arraysize]; //足される数の開始位置
    p2:=@buf1[convp-1]; //終了位置
    p3:=@buf2[arraysize]; //足す数の開始位置
    {$ASMMODE intel}  //高速化のためアセンブラで記述
    asm
      mov EDI,p1 //ポインタをセット
      mov ESI,p3
      clc //最初は桁上げなし
      lahf
  @DIV1LP:
      sahf
      mov EAX,[ESI]
      adc [EDI],EAX //キャリー付きの足し算、結果は同じ場所に格納
      lahf
      sub ESI,4 //ポインターを移動
      sub EDI,4
      cmp EDI,p2
      jge @DIV1LP
    end ['EAX','ESI','EDI']; //使用したレジスタをFPCに通知
end;

procedure sub(var buf1,buf2: buftyp; convp:integer);
//buf1からbuf2を引く。結果はbuf1に格納。convpは収束ポインタ、これより上位はオールゼロ
//引き算は下位から上位に処理
var
  p1,p2,p3  : pointer;
begin
    p1:=@buf1[arraysize]; //引かれる数の開始位置
    p2:=@buf1[convp-1]; //終了位置
    p3:=@buf2[arraysize]; //引く数の開始位置
    {$ASMMODE intel} //高速化のためアセンブラコードで記述
    asm
      mov EDI,p1
      mov ESI,p3
      clc //最初は借りなし
      lahf
  @DIV1LP:
      sahf
      mov EAX,[ESI]
      sbb [EDI],EAX //ボロー付きの引き算、結果は同じ配列に格納
      lahf
      sub ESI,4 //ポインターを移動
      sub EDI,4
      cmp EDI,p2
      jge @DIV1LP
    end ['EAX','ESI','EDI']; //使用したレジスタをFPCに通知
end;

procedure para_process(Index: PtrInt; Data: Pointer; Item: TMultiThreadProcItem);
//並列処理ルーチン
//計算式を4つに分割してそれぞれ計算
var
  n,convp        : integer;
  divbuf, midbuf : buftyp;

begin
  writeln('Thread ',index,' started');
  clearbuf(divbuf); //割り算用配列を初期化
  convp:=1; {initial value convergence pointer}
  n:=1; //係数を初期化
  case index of
   1 : begin
         divbuf[0]:=48;
         divide1(divbuf,18,convp);
         resbuf[index]:=divbuf;  //結果配列を初期化
         while convp<arraysize do
         begin
           n:=n+4; //係数
           divide1(divbuf,104976,convp); // 18^4
           divide2(divbuf,midbuf,n,convp);
           add(resbuf[index],midbuf,convp)
         end;
       end;
   2 : begin
         divbuf[0]:=48;
         divide1(divbuf,5832,convp); //3x18
         n:=3;
         divide2(divbuf,midbuf,n,convp);
         resbuf[index]:=midbuf;  //結果配列を初期化
         while convp<arraysize do
         begin
           n:=n+4; //係数
           divide1(divbuf,104976,convp); // 18^4
           divide2(divbuf,midbuf,n,convp);
           add(resbuf[index],midbuf,convp)
         end;
       end;
   3 : begin
         divbuf[0]:=32;
         divide1(divbuf,57,convp);
         resbuf[index]:=divbuf;  //結果配列を初期化
         while convp<arraysize do
         begin
           n:=n+2; //係数
           divide1(divbuf,3249,convp); // 57^2
           divide2(divbuf,midbuf,n,convp);
           if ((n-1) mod 4)=0 then add(resbuf[index],midbuf,convp)
                              else sub(resbuf[index],midbuf,convp);
         end;
       end;
   4 : begin
         divbuf[0]:=20;
         divide1(divbuf,239,convp);
         resbuf[index]:=divbuf;  //結果配列を初期化
         while convp<arraysize do
         begin
           n:=n+2; //係数
           divide1(divbuf,57121,convp); // 239^2
           divide2(divbuf,midbuf,n,convp);
           if ((n-1) mod 4)=0 then add(resbuf[index],midbuf,convp)
                              else sub(resbuf[index],midbuf,convp);
         end;
       end;
  end;
  writeln('Thread ',index,' finished');
end;


{main}
var
  i         : integer;
  starttime : Tdatetime;
  st        : string;
  f         : text;

begin
  writeln('Pi calculator version 3.10 (C) 2023 Dek / multithread - parallel processing');
  writeln;
  repeat
    write('Digits to calculate (multiple of 1,000, upto 10,000,000) = ');
    readln(goaldigit); //計算する桁数 1000の倍数
  until ((goaldigit mod 1000)=0) and (goaldigit>=1000) and (goaldigit<=maxdigit);
  arraysize := goaldigit div 8; //計算誤差防止のため配列を20%大きく取る 8 = 9.63/1.2(+20%)
  //計算開始
  starttime:=now;
  writeln;
  writeln(FloatToStrF(goaldigit,ffNumber,6,0)+' digits of the PI calculation started...');
  //並列処理ルーチンを起動
  ProcThreadPool.DoParallel(@para_process,1,4,nil);
  //4つの結果を合わせる
  sub(resbuf[1],resbuf[2],1);
  add(resbuf[1],resbuf[3],1);
  sub(resbuf[1],resbuf[4],1);
  //固定小数点2進を10進に変換 resbuf[1][0]が少数点以上
  writeln('Calculation finished.');
  writeln('Now converting the result to decimal');
  //結果をファイルに保存
  assign(f,'pi.txt');
  rewrite(f);
  writeln(f,FloatToStrF(goaldigit,ffNumber,6,0)+' digits of the PI.  Caliculation time:'+TimetoStr(now-starttime));
  writeln(f,resbuf[1][0],'.');
  for i:=1 to goaldigit div 5 do
  begin
    resbuf[1][0]:=0;
    mul100k(resbuf[1],arraysize-(i div 2)); //10万倍して少数点以上に出た数をとる(32ビット変数に入るのが5桁)
    st:='00000'+inttostr(resbuf[1][0]);
    st:=copy(st,length(st)-4,5); //0を加えて5桁に揃える
    write(f,st);
    if (i mod 10)=0 then writeln(f) //50桁ごとに改行
                    else
    if (i mod 2)=0 then write(f,' '); //10桁ごとにスペース
  end;
  close(f);
  writeln('Finished successfully. Saved in pi.txt.');
end.

